//.          GPS module for "Virtual Town"project
//.                  Copyright (c) 2005 
//.                                                       
//.  Authors: PAV <AlxPonom@mail.ru>             
//.                                                       
//.  This program is free software under the GPL (>= v2)  
//.  Read the file COPYING coming with project for details 

namespace GPS
{
	using System;
	using System.Threading;
	using System.IO.Ports;

	/// <summary>
	/// GPS NMEA reading
	/// </summary>
	public class GPSReading
	{
		private string PortName = "COM1"; 
		private int LinkErrorsCount = 1;
		private Thread Processing;
		private bool ProcessingCancelled;
		private NmeaInterpreter NMEAInterpreter;
		private bool FixObtained = false;
		private string SatellitesInformation = "";
		private double HDOP;
		private double VDOP;
		private double PDOP;
		private double Latitude;
		private double Longitude;
		private double Speed;
		private double Bearing;
		private System.DateTime LocalTime;
		public int UpdateInterval;
		
		public GPSReading(string pPortName)
		{
			PortName = pPortName;
			NMEAInterpreter = new NmeaInterpreter();
			//. assign parser handlers
			NMEAInterpreter.SatelliteReceived += new NmeaInterpreter.SatellitesReceivedEventHandler(OnSatellitesReceived);
			NMEAInterpreter.FixObtained += new NmeaInterpreter.FixObtainedEventHandler(OnFixObtained);
			NMEAInterpreter.FixLost += new NmeaInterpreter.FixLostEventHandler(OnFixLost);
			NMEAInterpreter.HDOPReceived += new NmeaInterpreter.HDOPReceivedEventHandler(OnHDOPReceived);
			NMEAInterpreter.VDOPReceived += new NmeaInterpreter.VDOPReceivedEventHandler(OnVDOPReceived);
			NMEAInterpreter.PDOPReceived += new NmeaInterpreter.PDOPReceivedEventHandler(OnPDOPReceived);
			NMEAInterpreter.PositionReceived += new NmeaInterpreter.PositionReceivedEventHandler(OnPositionReceived);
			NMEAInterpreter.SpeedReceived += new NmeaInterpreter.SpeedReceivedEventHandler(OnSpeedReceived);
			NMEAInterpreter.BearingReceived += new NmeaInterpreter.BearingReceivedEventHandler(OnBearingReceived);
			NMEAInterpreter.DateTimeChanged += new NmeaInterpreter.DateTimeChangedEventHandler(OnTimeReceived);
			//.
			ThreadStart ThreadDelegate = new ThreadStart(Execute);
			Processing = new Thread(ThreadDelegate);
			ProcessingCancelled = false;
			UpdateInterval = 2000;
			Processing.Start();
		}

		public void Execute()
		{
			SerialPort Port = new SerialPort(PortName,4800,Parity.None,8,StopBits.One);
			Port.ReadTimeout = 3*UpdateInterval;
			Port.Open();
			try
			{
				string Sentence = "";
				while (!ProcessingCancelled)
				{
					try
					{
						Sentence = Port.ReadLine();
						if (Sentence == null) throw new Exception("link broken"); //. =>
						//.
						lock(this) {LinkErrorsCount = 0;};
						try
						{
							NMEAInterpreter.Parse(Sentence);
						}
						catch {};
					}
					catch
					{
						lock(this) {LinkErrorsCount++;};
						//. try re-open
						Port.Close();
						Port.Open();
					};
				};
			}
			finally
			{
				Port.Close();
			};
		}

		public void Terminate()
		{
			ProcessingCancelled = true;
			Processing.Join(3*UpdateInterval);
		}


		public int LinkErrors()
		{
			lock(this) { return LinkErrorsCount; };
		}

		private void OnSatellitesReceived(string SatellitesInfo)
		{
			lock(this)
			{
				SatellitesInformation = SatellitesInfo;
			};
		}

		private void OnFixObtained()
		{
			lock(this)
			{
				FixObtained = true;
			}
		}

		private void OnFixLost()
		{
			lock(this)
			{
				FixObtained = false;
			};
		}

		private void OnHDOPReceived(double value)
		{
			lock(this)
			{
				HDOP = value;
			};
		}

		private void OnVDOPReceived(double value)
		{
			lock(this)
			{
				VDOP = value;
			};
		}

		private void OnPDOPReceived(double value)
		{
			lock(this)
			{
				PDOP = value;
			};
		}

		private void OnPositionReceived(double latitude, double longitude)
		{
			lock(this)
			{
				Latitude = latitude;
				Longitude = longitude;
			};
		}

		private void OnSpeedReceived(double speed)
		{
			lock(this)
			{
				Speed = speed;
			};
		}

		private void OnBearingReceived(double bearing)
		{
			lock(this)
			{
				Bearing = bearing;
			};
		}

		private void OnTimeReceived(System.DateTime dateTime)
		{
			lock(this)
			{
				LocalTime = dateTime;
			};
		}

		public bool IsFixObtained()
		{
			lock(this)
			{
				return FixObtained;
			};
		}

		public void GerParameters(out double oHDOP, out double oVDOP, out double oPDOP, out string oSatellitesInfo)
		{
			lock(this)
			{
				oHDOP = HDOP;
				oVDOP = VDOP;
				oPDOP = PDOP;
				oSatellitesInfo = SatellitesInformation;
			};
		}

		public void GetDATA(out double oLatitude, out double oLongitude, out double oSpeed, out double oBearing, out System.DateTime oLocalTime)
		{
			lock(this)
			{
				oLatitude = Latitude;
				oLongitude = Longitude;
				oSpeed = Speed;
				oBearing = Bearing;
				oLocalTime = LocalTime;
			};
		}
	}
}
